﻿using Microsoft.Extensions.Logging;
using SoftwareInstallationContracts.BindingModels;
using SoftwareInstallationContracts.BusinessLogicContracts;
using SoftwareInstallationContracts.SearchModels;
using SoftwareInstallationContracts.StoragesContracts;
using SoftwareInstallationContracts.ViewModels;
using SoftwareInstallationDataModels.Models;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace SoftwareInstallationBusinessLogic.BusinessLogics
{
    public class ShopLogic : IShopLogic
    {
        private readonly ILogger _logger;
        private readonly IShopStorage _ShopStorage;
        public ShopLogic(ILogger<ShopLogic> logger, IShopStorage ShopStorage)
        {
            _logger = logger;
            _ShopStorage = ShopStorage;
        }
        public bool AddPackage(ShopSearchModel model, IPackageModel package, int quantity)
        {
            if (model == null)
            {
                throw new ArgumentNullException(nameof(model));
            }

            if (quantity <= 0)
            {
                throw new ArgumentException("Количество добавляемого изделия должно быть больше 0", nameof(quantity));
            }

            _logger.LogInformation("AddPackageInShop. ShopName:{ShopName}.Id:{Id}", model.ShopName, model.Id);
            var element = _ShopStorage.GetElement(model);

            if (element == null)
            {
                _logger.LogWarning("AddPackageInShop element not found");
                return false;
            }

            if (element.PackageMaxCount - element.ShopPackages.Select(x => x.Value.Item2).Sum() < quantity)
            {
                throw new ArgumentNullException("Магазин переполнен", nameof(quantity));
            }

            _logger.LogInformation("AddPackageInShop find. Id:{Id}", element.Id);
            if (element.ShopPackages.TryGetValue(package.Id, out var pair))
            {
                element.ShopPackages[package.Id] = (package, quantity + pair.Item2);
                _logger.LogInformation("AddPackageInShop. Has been added {quantity} {package} in {ShopName}", quantity, package.PackageName, element.ShopName);
            }
            else
            {
                element.ShopPackages[package.Id] = (package, quantity);
                _logger.LogInformation("AddPackageInShop. Has been added {quantity} new Package {package} in {ShopName}", quantity, package.PackageName, element.ShopName);
            }

            _ShopStorage.Update(new()
            {
                Id = element.Id,
                ShopAddress = element.ShopAddress,
                ShopName = element.ShopName,
                OpeningDate = element.OpeningDate,
                ShopPackages = element.ShopPackages,
                PackageMaxCount = element.PackageMaxCount
            });
            return true;
        }
        public bool AddPackage(IPackageModel package, int quantity)
        {
            if (package == null)
            {
                throw new ArgumentNullException(nameof(package));
            }

            if (quantity <= 0)
            {
                throw new ArgumentException("Количество добавляемого изделия должно быть больше 0", nameof(quantity));
            }

            var freePlaces = _ShopStorage.GetFullList()
                .Select(x => x.PackageMaxCount - x.ShopPackages
                .Select(p => p.Value.Item2).Sum()).Sum() - quantity;

            if (freePlaces < 0)
            {
                _logger.LogInformation("AddPackage. Failed to add package to store. It's full.");
                return false;
            }

            foreach (var store in _ShopStorage.GetFullList())
            {
                var temp = Math.Min(quantity, store.PackageMaxCount - store.ShopPackages.Select(x => x.Value.Item2).Sum());

                if (temp <= 0)
                {
                    continue;
                }

                if (!AddPackage(new() { Id = store.Id }, package, temp))
                {
                    _logger.LogWarning("An error occurred while adding package to stores");
                    return false;
                }

                quantity -= temp;

                if (quantity == 0)
                {
                    return true;
                }
            }
            return true;
        }


        public bool Create(ShopBindingModel model)
        {
            CheckModel(model);
            model.ShopPackages = new();

            if (_ShopStorage.Insert(model) == null)
            {
                _logger.LogWarning("Insert operation failed");
                return false;
            }

            return true;
        }

        public bool SellPackage(IPackageModel package, int quantity)
        {
            return _ShopStorage.SellPackage(package, quantity);
        }

        public bool Delete(ShopBindingModel model)
        {
            CheckModel(model, false);
            _logger.LogInformation("Delete. Id:{Id}", model.Id);

            if (_ShopStorage.Delete(model) == null)
            {
                _logger.LogWarning("Delete operation failed");
                return false;
            }

            return true;
        }

        public ShopViewModel? ReadElement(ShopSearchModel model)
        {
            if (model == null)
            {
                throw new ArgumentNullException(nameof(model));
            }

            _logger.LogInformation("ReadElement. ShopName:{ShopName}.Id:{ Id}", model.ShopName, model.Id);
            var element = _ShopStorage.GetElement(model);

            if (element == null)
            {
                _logger.LogWarning("ReadElement element not found");
                return null;
            }

            _logger.LogInformation("ReadElement find. Id:{Id}", element.Id);
            return element;
        }

        public List<ShopViewModel>? ReadList(ShopSearchModel? model)
        {
            _logger.LogInformation("ReadList. ShopName:{ShopName}.Id:{ Id} ", model?.ShopName, model?.Id);

            var list = (model == null) ? _ShopStorage.GetFullList() : _ShopStorage.GetFilteredList(model);

            if (list == null)
            {
                _logger.LogWarning("ReadList return null list");
                return null;
            }

            _logger.LogInformation("ReadList. Count:{Count}", list.Count);
            return list;
        }

        public bool Update(ShopBindingModel model)
        {
            CheckModel(model, false);

            if (string.IsNullOrEmpty(model.ShopName))
            {
                throw new ArgumentNullException("Нет названия магазина", nameof(model.ShopName));
            }

            if (_ShopStorage.Update(model) == null)
            {
                _logger.LogWarning("Update operation failed");
                return false;
            }
            return true;
        }

        private void CheckModel(ShopBindingModel model, bool withParams = true)
        {
            if (model == null)
            {
                throw new ArgumentNullException(nameof(model));
            }

            if (!withParams)
            {
                return;
            }

            if (string.IsNullOrEmpty(model.ShopName))
            {
                throw new ArgumentNullException("Нет названия магазина", nameof(model.ShopName));
            }

            if (model.PackageMaxCount < 0)
            {
                throw new ArgumentException("Максимальное количество изделий в магазине не может быть меньше нуля", nameof(model.PackageMaxCount));
            }

            _logger.LogInformation("Shop. ShopName:{0}.ShopAddress:{1}. Id: {2}", model.ShopName, model.ShopAddress, model.Id);
            var element = _ShopStorage.GetElement(new ShopSearchModel
            {
                ShopName = model.ShopName
            });

            if (element != null && element.Id != model.Id && element.ShopName == model.ShopName)
            {
                throw new InvalidOperationException("Магазин с таким названием уже есть");
            }
        }
    }
}
